@@ -1,25 +1,28 @@
 """Client library for interacting with Hydrawise's cloud API."""
 
-import logging
 from datetime import datetime
 from functools import cache
+from importlib import resources
+import logging
 
 from apischema.graphql import graphql_schema
 from gql import Client
 from gql.dsl import DSLField, DSLMutation, DSLQuery, DSLSchema, DSLSelectable, dsl_gql
-from gql.transport.aiohttp import AIOHTTPTransport
-from gql.transport.aiohttp import log as gql_log
-from graphql import GraphQLSchema
+from gql.transport.aiohttp import AIOHTTPTransport, log as gql_log
+from graphql import GraphQLSchema, build_ast_schema, parse
 
 from .auth import Auth
+from .base import HydrawiseBase
 from .exceptions import MutationError
 from .schema import (
     Controller,
     DateTime,
-    Mutation,
+    LocalizedValueType,
+    Sensor,
+    SensorFlowSummary,
     StatusCodeAndSummary,
-    Query,
     User,
+    WateringReportEntry,
     Zone,
     ZoneSuspension,
 )
@@ -33,13 +36,11 @@ API_URL = "https://app.hydrawise.com/api/v2/graph"
 
 @cache
 def _get_schema() -> GraphQLSchema:
-    return graphql_schema(
-        query=[getattr(Query, m) for m in Query.__abstractmethods__],
-        mutation=[getattr(Mutation, m) for m in Mutation.__abstractmethods__],
-    )
+    schema_text = resources.files(__package__).joinpath("hydrawise.graphql").read_text()
+    return build_ast_schema(parse(schema_text))
 
 
-class Hydrawise:
+class Hydrawise(HydrawiseBase):
     """Client library for interacting with Hydrawise sprinkler controllers.
 
     Should be instantiated with an Auth object that handles authentication and low-level transport.
@@ -74,12 +75,16 @@ class Hydrawise:
                 # Assume bool response
                 raise MutationError
 
-    async def get_user(self) -> User:
+    async def get_user(self, fetch_zones: bool = True) -> User:
         """Retrieves the currently authenticated user.
 
+        :param fetch_zones: Not used in this implementation.
         :rtype: User
         """
-        selector = self._schema.Query.me.select(*get_selectors(self._schema, User))
+        skip = [] if fetch_zones else ["controllers.zones"]
+        selector = self._schema.Query.me.select(
+            *get_selectors(self._schema, User, skip)
+        )
         result = await self._query(selector)
         return deserialize(User, result["me"])
 
@@ -137,7 +142,7 @@ class Hydrawise:
         zone: Zone,
         mark_run_as_scheduled: bool = False,
         custom_run_duration: int = 0,
-    ):
+    ) -> None:
         """Starts a zone's run cycle.
 
         :param zone: The zone to start.
@@ -157,7 +162,7 @@ class Hydrawise:
         )
         await self._mutation(selector)
 
-    async def stop_zone(self, zone: Zone):
+    async def stop_zone(self, zone: Zone) -> None:
         """Stops a zone.
 
         :param zone: The zone to stop.
@@ -172,7 +177,7 @@ class Hydrawise:
         controller: Controller,
         mark_run_as_scheduled: bool = False,
         custom_run_duration: int = 0,
-    ):
+    ) -> None:
         """Starts all zones attached to a controller.
 
         :param controller: The controller whose zones to start.
@@ -192,7 +197,7 @@ class Hydrawise:
         )
         await self._mutation(selector)
 
-    async def stop_all_zones(self, controller: Controller):
+    async def stop_all_zones(self, controller: Controller) -> None:
         """Stops all zones attached to a controller.
 
         :param controller: The controller whose zones to stop.
@@ -204,7 +209,7 @@ class Hydrawise:
         )
         await self._mutation(selector)
 
-    async def suspend_zone(self, zone: Zone, until: datetime):
+    async def suspend_zone(self, zone: Zone, until: datetime) -> None:
         """Suspends a zone's schedule.
 
         :param zone: The zone to suspend.
@@ -218,7 +223,7 @@ class Hydrawise:
         )
         await self._mutation(selector)
 
-    async def resume_zone(self, zone: Zone):
+    async def resume_zone(self, zone: Zone) -> None:
         """Resumes a zone's schedule.
 
         :param zone: The zone whose schedule to resume.
@@ -228,7 +233,7 @@ class Hydrawise:
         )
         await self._mutation(selector)
 
-    async def suspend_all_zones(self, controller: Controller, until: datetime):
+    async def suspend_all_zones(self, controller: Controller, until: datetime) -> None:
         """Suspends the schedule of all zones attached to a given controller.
 
         :param controller: The controller whose zones to suspend.
@@ -242,7 +247,7 @@ class Hydrawise:
         )
         await self._mutation(selector)
 
-    async def resume_all_zones(self, controller: Controller):
+    async def resume_all_zones(self, controller: Controller) -> None:
         """Resumes the schedule of all zones attached to the given controller.
 
         :param controller: The controller whose zones to resume.
@@ -254,7 +259,7 @@ class Hydrawise:
         )
         await self._mutation(selector)
 
-    async def delete_zone_suspension(self, suspension: ZoneSuspension):
+    async def delete_zone_suspension(self, suspension: ZoneSuspension) -> None:
         """Removes a specific zone suspension.
 
         Useful when there are multiple suspensions for a zone in effect.
@@ -265,3 +270,103 @@ class Hydrawise:
             id=suspension.id
         ).select()
         await self._mutation(selector)
+
+    async def get_sensors(self, controller: Controller) -> list[Sensor]:
+        """Retrieves sensors associated with the given controller.
+
+        :param controller: Controller whose sensors to fetch.
+        :rtype: list[Sensor]
+        """
+        selector = self._schema.Query.controller(controllerId=controller.id).select(
+            self._schema.Controller.sensors.select(
+                *get_selectors(self._schema, Sensor)
+            ),
+        )
+
+        result = await self._query(selector)
+        return deserialize(list[Sensor], result["controller"]["sensors"])
+
+    async def get_water_flow_summary(
+        self, controller: Controller, sensor: Sensor, start: datetime, end: datetime
+    ) -> SensorFlowSummary:
+        """Retrieves the water flow summary for a given sensor.
+
+        :param controller: Controller that controls the sensor.
+        :param sensor: Sensor for which a water flow summary is fetched.
+        :param start:
+        :param end:
+        :rtype: list[Sensor]
+        """
+        selector = self._schema.Query.controller(controllerId=controller.id).select(
+            self._schema.Controller.sensors.select(
+                *get_selectors(self._schema, Sensor),
+                self._schema.Sensor.flowSummary(
+                    start=DateTime.to_json(start).timestamp,
+                    end=DateTime.to_json(end).timestamp,
+                ).select(*get_selectors(self._schema, SensorFlowSummary)),
+            ),
+        )
+
+        result = await self._query(selector)
+
+        # There is no way to query for one particular sensor through GraphQL. We need
+        # to filter here instead. This should not really be a performance problem in
+        # practice since it is unlikely for a controller to have more than one water
+        # sensor.
+        sensors = list(
+            filter(lambda s: s["id"] == sensor.id, result["controller"]["sensors"])
+        )
+        if len(sensors) == 0:
+            raise ValueError(f"Sensor with id={sensor.id} not found")
+        if "flowSummary" not in sensors[0]:
+            raise ValueError(
+                f"Sensor with id={sensor.id} does not have any flow information"
+            )
+        if sensors[0]["flowSummary"] is None:
+            return SensorFlowSummary(total_water_volume=LocalizedValueType(0.0, "gal"))
+        return deserialize(SensorFlowSummary, sensors[0]["flowSummary"])
+
+    async def get_watering_report(
+        self, controller: Controller, start: datetime, end: datetime
+    ) -> list[WateringReportEntry]:
+        """Retrieves a watering report for the given controller and time period.
+
+        :param controller: The controller whose watering report to generate.
+        :param start: Start time.
+        :param end: End time."""
+        selector = self._schema.Query.controller(controllerId=controller.id).select(
+            self._schema.Controller.reports.select(
+                self._schema.Reports.watering(
+                    **{
+                        "from": int(start.timestamp()),
+                        "until": int(end.timestamp()),
+                    }
+                ).select(*get_selectors(self._schema, WateringReportEntry)),
+            ),
+        )
+        result = await self._query(selector)
+        entries = deserialize(
+            list[WateringReportEntry], result["controller"]["reports"]["watering"]
+        )
+        # The call to watering() can return events outside of the provided time interval.
+        # Filter out events that happen before or after the provided time interval.
+        return list(
+            filter(
+                lambda entry: entry.run_event is not None
+                and entry.run_event.reported_start_time is not None
+                and entry.run_event.reported_end_time is not None
+                and (
+                    (
+                        start.timestamp()
+                        <= entry.run_event.reported_start_time.timestamp()
+                        <= end.timestamp()
+                    )
+                    or (
+                        start.timestamp()
+                        <= entry.run_event.reported_end_time.timestamp()
+                        <= end.timestamp()
+                    )
+                ),
+                entries,
+            )
+        )
